;; Generated package description from evil-nerd-commenter.el  -*- no-byte-compile: t -*-
(define-package "evil-nerd-commenter" "3.6.1" "Comment/uncomment lines efficiently. Like Nerd Commenter in Vim" '((emacs "26.1")) :commit "3b197a2b559b06a7cf39978704b196f53dac802a" :authors '(("Chen Bin" . "chenbin.sh@gmail.com")) :maintainer '("Chen Bin" . "chenbin.sh@gmail.com") :keywords '("convenience" "evil") :url "http://github.com/redguardtoo/evil-nerd-commenter")
